public class ArithCoder
{
    private Bitstream BS;
    
    // Konstanten zur Bereichsunterteilung bei 32-Bit-Integern 
    // oberstes Bit wird zur Vermeidung von berlufen freigehalten
    private int g_FirstQuarter = 0x20000000;
    private int g_ThirdQuarter = 0x60000000;
    private int g_Half         = 0x40000000;

    private int mBitCount;
    
    // fuer Encoder und Decoder
    private int mLow;
    private int mHigh; // nur untere 31 bit um Ueberlauf zu vermeiden
    private int mStep;
    private int mScale;
    
    // fuer Decoder
    private int mBuffer;

    private boolean bitBool;

    public ArithCoder (Bitstream BS)
    {
	this.BS = BS;
	mBitCount = 0;
	mLow = 0;
	mHigh = 0x7FFFFFFF; // nur untere 31 bit um Ueberlauf zu vermeiden
	mStep = 0;
	mScale = 0;
    
	mBuffer = 0;	
    }

    private void SetBit(int bit) throws java.io.IOException
    {
        if (bit == 1) bitBool = true;
	else bitBool = false;
	BS.WriteBit(bitBool);
	System.out.println("Hallo" + mBitCount + " Bit: " + bit);
	mBitCount++;
	mBitCount %= 8;
    }
    
    private void SetBitFlush() throws java.io.IOException
    //Puffer bis zur naechsten Byte-Grenze mit Nullen fuellen. 
    {
	while (mBitCount != 0) SetBit(0);
    }

    private int GetBit() throws java.io.EOFException, java.io.IOException
    {
	bitBool = BS.ReadBit();
	if (bitBool == true) return 1;
	return 0;
    }

    private void Encode (int low_count, int high_count, int total) throws java.io.IOException
    {
	// Bereich in Schritte unterteilen
	mStep = ( mHigh - mLow + 1 ) / total; // oben offenes intervall => +1

	// obere Grenze aktualisieren
	mHigh = mLow + mStep * high_count - 1; // oben offenes intervall => -1
	
	// untere Grenze aktualisieren
	mLow = mLow + mStep * low_count;

	// e1/e2 Mapping durchfhren
	while( ( mHigh < g_Half ) || ( mLow >= g_Half ) )
	{
	    if( mHigh < g_Half )
	    {
		SetBit(0);
		mLow = mLow * 2;
		mHigh = mHigh * 2 + 1;

		// e3
		for(; mScale > 0; mScale-- ) SetBit(1);
	    }
	    else if( mLow >= g_Half )
	    {
		SetBit(1);
		mLow = 2 * ( mLow - g_Half );
		mHigh = 2 * ( mHigh - g_Half ) + 1;

		// e3
		for(; mScale > 0; mScale-- ) SetBit(0);
	    }
	}

	// e3
	while( ( g_FirstQuarter <= mLow ) && ( mHigh < g_ThirdQuarter ) )
	{
		mScale++;
		mLow = 2 * ( mLow - g_FirstQuarter );
		mHigh = 2 * ( mHigh - g_FirstQuarter ) + 1;
	}

    }

    private void EncodeFinish() throws java.io.IOException
    {
	// Es gibt zwei Mglichkeiten, wie mLow und mHigh liegen, d.h.
	// zwei Bits reichen zur Entscheidung aus.
	
	if( mLow < g_FirstQuarter ) // mLow < FirstQuarter < Half <= mHigh
	{
	    SetBit(0);

	    for( int i=0; i<mScale+1; i++ ) // 1 + e3-Skalierung abbauen
	       SetBit(1);
	}
	else // mLow < Half < ThirdQuarter <= mHigh
	{
	    SetBit(1); // der Decoder fgt die Nullen automatisch an
	}

	// Ausgabepuffer leeren
	SetBitFlush();
    }

    private void DecodeStart() throws java.io.EOFException, java.io.IOException
    {
	// Puffer mit Bits aus dem Eingabe-Code fllen
	for( int i=0; i<31; i++ ) // benutze nur die unteren 31 bit
	    mBuffer = ( mBuffer << 1 ) | GetBit();
    }
    
    public int DecodeTarget (int total)
    {
	// Bereich in Schritte unterteilen
	mStep = ( mHigh - mLow + 1 ) / total; // oben offenes intervall => +1

	// aktuellen Wert zurckgeben
	return ( mBuffer - mLow ) / mStep;	
    }

    private void Decode (int low_count, int high_count, int total) throws java.io.EOFException, java.io.IOException
    {
	// obere Grenze aktualisieren
	mHigh = mLow + mStep * high_count - 1; // oben offenes intervall => -1

	// untere Grenze aktualisieren
	mLow = mLow + mStep * low_count;

	// e1/e2
	while( ( mHigh < g_Half ) || ( mLow >= g_Half ) )
	{
	    if( mHigh < g_Half )
	    {
		mLow = mLow * 2;
		mHigh = mHigh * 2 + 1;
		mBuffer = 2 * mBuffer + GetBit();
	    }
	    else if( mLow >= g_Half )
	    {
		mLow = 2 * ( mLow - g_Half );
		mHigh = 2 * ( mHigh - g_Half ) + 1;
		mBuffer = 2 * ( mBuffer - g_Half ) + GetBit();
	    }
	    mScale = 0;
	}

	// e3
	while( ( g_FirstQuarter <= mLow ) && ( mHigh < g_ThirdQuarter ) )
	{
	    mScale++;
	    mLow = 2 * ( mLow - g_FirstQuarter );
	    mHigh = 2 * ( mHigh - g_FirstQuarter ) + 1;
	    mBuffer = 2 * ( mBuffer - g_FirstQuarter ) + GetBit();
	}
	
    }
    

    public static void main(String[] args)
    {
	try
	{
	    Bitstream BS = new Bitstream("suppe.dat");
	    ArithCoder mArithCoder = new ArithCoder(BS);
	    
	    int total = 100;

	    // mArithCoder.Encode(low_count, high_count, total);	    
	    mArithCoder.Encode(0, 20, total);
	    mArithCoder.Encode(70, 100, total);
	    mArithCoder.Encode(20, 70, total);
	    mArithCoder.EncodeFinish();
	    BS.Close();
	   
	    BS = new Bitstream("suppe.dat");
	    mArithCoder = new ArithCoder(BS);

	    mArithCoder.DecodeStart();
	    int value = mArithCoder.DecodeTarget(total);
	    System.out.println(value);
	    mArithCoder.Decode(0, 20, total);
	    BS.Close();
	}
	catch (java.io.FileNotFoundException e) 
	{
            System.out.println("File not found:");
            e.printStackTrace();
        } 
	catch (java.io.EOFException e) 
	{
            System.out.println("EOFException:");
            e.printStackTrace();
        } 
	catch (java.io.IOException e) 
	{
            System.out.println("IOException:");
            e.printStackTrace();
        }
    }

} // end class
