import java.util.LinkedList;

class BurrowsWheeler {

    private static final int END_MARKER = Integer.MAX_VALUE;

    // Sortiert die Zeilen in einer data.Length()x data.Length() Matrix lexikographisch, indem
    // der ASCII Wert der einzelnen Zeichen verglichen wird
    private static void sortByASCII(LinkedList rotationTable, int numberOfColumns) {
        LinkedList nextListToSet, listToCompare;
        boolean setList = false;
        int nextListToSetIndex;

        // Auesserste Schleife: Sucht sich fuer jede Zeile die noch "niedrigste" (niedrig im Kontext der
        // int Werte der ASCII Zeichen am Anfang der Zeile) Zeile und verschiebt die an die momentane Position (x)
        for (int x = 0; x < rotationTable.size(); x++) {
            nextListToSet = (LinkedList) rotationTable.get(x);
            nextListToSetIndex = x;

            // alle Zeilen nach der x-ten mit der x-ten vergleichen
            for (int y = (x + 1); y < rotationTable.size(); y++) {
                listToCompare = (LinkedList) rotationTable.get(y);

                // zeichenweise vergleichen; aufhoeren, wenn zeichen ungleich oder die beiden zeilen ident
                for (int z = 0; z < numberOfColumns; z++) {
                    if (((Integer) nextListToSet.get(z)).intValue() > ((Integer) listToCompare.get(z)).intValue()) {
                        setList = false;
                        break;
                    } else if (((Integer) nextListToSet.get(z)).intValue() < ((Integer) listToCompare.get(z)).intValue()) {
                        setList = true;
                        break;
                    }
                    if (z + 1 == rotationTable.size())
                        setList = true;
                }
                // Nur wenn der Anfang der momentanen Vergleichszeile lexikographisch "niedriger" war wird die
                // Liste mitsamt ihrem Index als momentan kleinste Zeile akzeptiert
                if (!setList) {
                    nextListToSet = listToCompare;
                    nextListToSetIndex = y;
                }
            }
            // Entfernt die lexikographisch niedrigste List von ihrer Position und fuegt sie an die momentane
            // Position in der Rotationstabelle ein
            rotationTable.add(x, rotationTable.remove(nextListToSetIndex));
        }
    }

    // Fuehrt eine Burrows Wheeler Transformation auf dem Daten Array aus
    // gibt die Position des Endemarkers zurck
    static int Transform (Int1DArray data) {
        // Achtung: mein Rotationstabelle ist kein 2 dimensionales Array sondern ein linked list von linked lists
        // Dies begruendet sich damit, das man in Java keine so schoenen expliziten Pointer wie in (C/C++) hat und damit
        // das herumkopieren von ganzen Zeilen in 2 dimensionalen Arrays viel mehr Aufwand waere.
        LinkedList rotationTable = new LinkedList();
        LinkedList tempList;

        // Erzeugung der Rotationstabelle (linked list von linked lists)
        // und gleichzeitige Auffuellung mit den Daten
        for (int x = 0; x < data.Length() + 1; x++) {
            tempList = new LinkedList();
            for (int y = x; y < (data.Length() + x); y++)
                tempList.addLast(new Integer(data.Get(y % data.Length())));
            tempList.add (data.Length () - x, new Integer (END_MARKER));
            rotationTable.addLast(tempList);
        }

        // Sortierung der Eintraege in der Rotationstabelle
        sortByASCII(rotationTable, rotationTable.size());

        // Eintragen der letzten Spalte ins Daten array
        int y = 0;
        int MarkerPos = 0;
        int v;
        for (int x = 0; x < data.Length(); x++) {
            tempList = (LinkedList) rotationTable.get(x);
            v = ((Integer) tempList.getLast()).intValue();
            if   (v == END_MARKER)
                 MarkerPos = x;
            else data.Set(y++, v);
        }

        return MarkerPos;
    }

    // Transformiert das Daten Array in seine urspruengliche Form zurueck
    static void InverseTransform (Int1DArray data, int MarkerPos) {
        LinkedList rotationTable = new LinkedList();
        LinkedList tempList;

        // Erzeugen der initalen Rotationsmatrix und anschliessende erste Sortierung
        for (int x = 0; x < data.Length(); x++) {
            if (x == MarkerPos)
            {
               tempList = new LinkedList();
               tempList.addLast(new Integer(END_MARKER));
               rotationTable.addLast(tempList);
            }
            tempList = new LinkedList();
            tempList.addLast(new Integer(data.Get(x)));
            rotationTable.addLast(tempList);
        }
        sortByASCII(rotationTable, 1);

        // Hinzufuegen und sortieren von Eintraegen
        for (int x = 1; x < data.Length() + 1; x++) {
            int z = 0;
            for (int y = 0; y < data.Length(); y++) {
                tempList = (LinkedList) rotationTable.get(z);
                if (y == MarkerPos) {
                   tempList.addFirst (new Integer(END_MARKER));
                   z ++;
                   tempList = (LinkedList) rotationTable.get(z);
                }
                tempList.addFirst(new Integer(data.Get(y)));
                z ++;
            }
            sortByASCII(rotationTable, x + 1);
        }

        // Heraussuchen der Zeile mit der Endekennung am Schluss
        // und schreiben dieser Zeile ins Daten array
        for (int x = 0; x < data.Length() + 1; x++) {
            tempList = (LinkedList) rotationTable.get(x);
            if (((Integer) tempList.getLast()).intValue() == END_MARKER) {
                for (int y = 0; y < data.Length(); y++) {
                    data.Set(y, ((Integer) tempList.get(y)).intValue());
                }
            }
        }
    }

    // Macht ein simples Runlength Coding auf dem Daten Array
    // War nicht meine Aufgabe; Hab da nur ein bisserl rumgespielt und es dringelassen
    private static void doRunlengthCoding(Int1DArray data) {
        int offset = 0;
        int offset2 = 0;
        String symbol = "";
        LinkedList tempData = new LinkedList();
        LinkedList compressedData = new LinkedList();

        for (int x = 0; x < data.Length(); x++) {
            if (x == 0) {
                if (data.Get(x) == data.Get(x + 1)) {
                    symbol = "" + (char) data.Get(0);
                    offset2++;
                } else {
                    tempData.addLast("" + (char) data.Get(0));
                    offset++;
                }
            } else if (x + 1 == data.Length()) {
                if (data.Get(x) == data.Get(x - 1)) {
                    compressedData.addLast("R"+ (offset2+1));
                    compressedData.addLast(symbol);
                } else {
                    offset++;
                    tempData.addLast("" +(char) data.Get(x));
                    compressedData.addLast("N"+ offset);
                    while (tempData.size() > 0)
                        compressedData.addLast(tempData.removeFirst());
                }
            } else if (data.Get(x) == data.Get(x - 1)) {
                offset2++;
                if(data.Get(x) != data.Get(x + 1)) {
                    compressedData.addLast("R"+ offset2);
                    compressedData.addLast(symbol);
                    offset2 = 0;
                }
            } else if (data.Get(x) != data.Get(x - 1)) {
                if (data.Get(x) == data.Get(x + 1)) {
                    compressedData.addLast("N"+ offset);
                    while (tempData.size() > 0)
                        compressedData.addLast(tempData.removeFirst());
                    offset = 0;
                    symbol = "" + (char) data.Get(x);
                    offset2++;
                } else {
                    tempData.addLast("" + (char) data.Get(x));
                    offset++;
                }

            }
        }

        System.out.print("("+((String) compressedData.get(0)));
        for (int x = 1; x < compressedData.size(); x++) {
            System.out.print(","+(String) compressedData.get(x));
        }
        System.out.print(")");
    }

    private static void PrintTable (LinkedList T)
    {
        System.out.println ("\n");
        for (int x = 0; x < T.size (); x ++)
        { for (int y = 0; y < T.size (); y ++)
          {  int v = ((Integer) ((LinkedList) T.get(x)).get(y)).intValue ();
             char c = (char) v;
             if (v == END_MARKER) c = '.';
             System.out.print (c);
          }
          System.out.println ();
        }
    }

    public static void main(String args[]) {
        Quantized1D data;
        if   (args.length == 0)
             data = new Quantized1D("ADAADAADAADAADA");
        else data = new Quantized1D(args[0]);

        System.out.println("\nDaten vor der Transformation:");
        data.PrintChars ();

        // Durchfuehrung der BWT
        int m = BurrowsWheeler.Transform (data);
        System.out.println("\nDaten nach der Transformation:");
        data.PrintChars ();
        System.out.print ("\nMarker-Position: ");
        System.out.println (m);

        // Runlength Coding
        System.out.println("\nErgebnis eines simplen Runlength Coding:");
        doRunlengthCoding(data);
        System.out.println("");

        // Durchfuehrung der inversen BWT
        BurrowsWheeler.InverseTransform (data, m);
        System.out.println("\nDaten ruecktransformiert:");
        data.PrintChars ();
        System.out.println("");
    }
}
