class Filter extends Double1DArray
{
  private double Coeffs[];

  private int Offset;

  private int Downsampling;

  Filter (int CoeffCount, int Offset, int Downsampling)
  {
    Coeffs = new double[CoeffCount];
    this.Offset = Offset;
    this.Downsampling = Downsampling;
  }

  public double Get (int n)
  {
    return Coeffs[n];
  }

  public void Set (int n, double v)
  {
    Coeffs[n] = v;
  }

  public int Length ()
  {
    return Coeffs.length;
  }

  static Filter NewLowHaar ()
  {
    Filter F = new Filter (2, 0, 2);
    F.Set (0, Math.sqrt (0.5));
    F.Set (1, Math.sqrt (0.5));
    return F;
  }

  static Filter NewHighHaar ()
  {
    Filter F = new Filter (2, 0, 2);
    F.Set (0, Math.sqrt (0.5));
    F.Set (1, - Math.sqrt (0.5));
    return F;
  }

  static Filter NewLowDaub4 ()
  {
    Filter F = new Filter (4, -1, 2);
    F.Set (0, 0.482963);
    F.Set (1, 0.836516);
    F.Set (2, 0.224144);
    F.Set (3, -0.12941);
    return F;
  }

  static Filter NewHighDaub4 ()
  {
    Filter F = new Filter (4, -1, 2);
    F.Set (0, -0.12941);
    F.Set (1, -0.224144);
    F.Set (2, 0.836516);
    F.Set (3, -0.482963);
    return F;
  }

  void Apply (Double1DArray From, Double1DArray To)
  {
    int k = 0;
    double sum = 0;
    int fromIndex = 0;
    for (int i = 0; i < To.Length (); i ++)
    {
      sum = 0;
      for(int j = 0; j < Length(); j ++)
      {
        fromIndex = (Offset + k + j);
        while (fromIndex < 0) fromIndex += From.Length();
        while (fromIndex >= From.Length()) fromIndex -= From.Length();

        sum += From.Get(fromIndex) * Get(j);
      }

      To.Set(i, sum);
      k += Downsampling;
    }        
  }

  void InverseApply (Double1DArray From, Double1DArray To)
  {
    int toIndex = 0;
    int k = 0;

    for(int i = 0; i < From.Length(); i++)
    {
      for(int j = 0; j < Length(); j ++)
      {
        toIndex = (Offset + k + j) % To.Length();
        while (toIndex < 0) toIndex += To.Length();
        while (toIndex > To.Length()) toIndex -= To.Length();

        To.Set(toIndex, To.Get(toIndex) + From.Get(i)*Get(j));
      }
      k += Downsampling;        
    }
  }

  public static void main (String args[])
  {
     Signal1D S = new Signal1D (8);
     S.Set (0, 1.5); S.Set (1, 2.5); S.Set (2, 3.5); S.Set (3, 3.5);
     S.Set (4, 4.5); S.Set (5, 4.5); S.Set (6, 3.5); S.Set (7, 2.5);
     System.out.print ("original: ");  S.Print ();

     Signal1D SL = new Signal1D (4);
     Filter L = NewLowDaub4 ();
     L.Apply (S, SL);
     System.out.print ("low-pass: ");  SL.Print ();

     Signal1D SH = new Signal1D (4);
     Filter H = NewHighDaub4 ();
     H.Apply (S, SH);
     System.out.print ("high-pass: ");  SH.Print ();

     S.SetAll (0);
     L.InverseApply (SL, S);
     System.out.print ("reconstructed low-pass: ");  S.Print ();
     H.InverseApply (SH, S);
     System.out.print ("reconstructed complete: ");  S.Print ();
  }
}
