class QuadTreeCoder
{
  private Double2DArray Image;
  private double Max;
  private Quantized2D Status;
  private EntropyCoder Coder;

  final static int INSIGNIFICANT = 0;
  final static int SIGNIFICANT = 1;
  final static int PARTIAL = 2;

  final static int NODE_CONTEXT = 0;
  final static int LEAF_CONTEXT = 1;
  final static int REFINEMENT_CONTEXT = 2;
  final static int SIGN_CONTEXT = 3;

  QuadTreeCoder (Double2DArray Image)
  {
    this.Image = Image;
  }

  void Encode (EntropyCoder Coder, int Bitplanes)
  {
    this.Coder = Coder;
    Max = Math.ceil (Image.AbsMax () + 0.01);
    Coder.WriteInt ((int) Max);
    Status = new Quantized2D (Image.Dimension ());
    double threshold = Max;
    for (int i = 0; i < Bitplanes; i ++)
    {
      threshold *= 0.5;
      EncodeBitplane (new Double2DSubArray (Image, new Position (0, 0), Image.Dimension ()), threshold, PARTIAL);
    }
  }

  void EncodeBitplane (Double2DSubArray A, double threshold, int ParentStatus)
  {
    //A.GetOrigin ().Print ();
    //A.Dimension ().Print ();
    //System.out.println (ParentStatus);

    int NewStatus = ParentStatus;

    if (ParentStatus == PARTIAL)
    {
      int NodeContext = NODE_CONTEXT;
      if (A.Size () == 1)
         NodeContext = LEAF_CONTEXT;

      int SigCount = 0;
      int InsigCount = 0;
      for (PositionIterator i = new PositionIterator (A.Dimension ()); ! i.AtEnd (); i.Iterate ())
      {
        if (Status.Get (A.GetOrigin ().Plus (i.Current ())) != SIGNIFICANT)
        {  if   (Math.abs (A.Get (i.Current ())) >= threshold)
                SigCount ++;
           else InsigCount ++;
        }
      }

      if (SigCount > 0)
      {  if (InsigCount > 0)
         {  Coder.Encode (PARTIAL, NodeContext);
            NewStatus = PARTIAL;
         }
         else
         {  Coder.Encode (SIGNIFICANT, NodeContext);
            NewStatus = SIGNIFICANT;
         }
      }
      else
      {  if (InsigCount > 0)
         {  Coder.Encode (INSIGNIFICANT, NodeContext);
            NewStatus = INSIGNIFICANT;
         }
         else
         {  NewStatus = INSIGNIFICANT; // just not PARTIAL
         }
      }
    }

    if (A.Size () > 1)
    {
       for (int i = 0; i < 4; i ++)
           EncodeBitplane (A.Quadrant (i), threshold, NewStatus);
    }
    else
    {
       Position p = A.GetOrigin ();
       if (Status.Get (p) == INSIGNIFICANT)
       {  if (NewStatus == SIGNIFICANT)
          {  Coder.Encode (Image.Get (p) > 0 ? 1 : 0, SIGN_CONTEXT);
             Status.Set (p, NewStatus);
          }
       }
       else
       {  double q = Math.abs (Image.Get (p)) / threshold;
          Coder.Encode (q - 2.0 * Math.floor (q / 2.0) >= 1.0 ? 1 : 0, REFINEMENT_CONTEXT);
       }
    }
  }

  void Decode (EntropyCoder Coder, int Bitplanes)
    throws java.io.IOException
  {
    this.Coder = Coder;
    Max = (double) Coder.ReadInt ();
    Status = new Quantized2D (Image.Dimension ());
    double threshold = Max;
    for (int i = 0; i < Bitplanes; i ++)
    {
      threshold *= 0.5;
      DecodeBitplane (new Double2DSubArray (Image, new Position (0, 0), Image.Dimension ()), threshold, PARTIAL);
    }
  }

  void DecodeBitplane (Double2DSubArray A, double threshold, int ParentStatus)
    throws java.io.IOException
  {
    //A.GetOrigin ().Print ();
    //A.Dimension ().Print ();
    //System.out.println (ParentStatus);

    int NewStatus = ParentStatus;

    if (ParentStatus == PARTIAL)
    {
      int NodeContext = NODE_CONTEXT;
      if (A.Size () == 1)
         NodeContext = LEAF_CONTEXT;

      int InsigCount = 0;
      for (PositionIterator i = new PositionIterator (A.Dimension ()); ! i.AtEnd (); i.Iterate ())
        if (Status.Get (A.GetOrigin ().Plus (i.Current ())) == INSIGNIFICANT)
           InsigCount ++;

      if (InsigCount > 0)
         NewStatus = Coder.Decode (NodeContext);
    }

    if (A.Size () > 1)
    {
       for (int i = 0; i < 4; i ++)
           DecodeBitplane (A.Quadrant (i), threshold, NewStatus);
    }
    else
    {
       Position p = A.GetOrigin ();
       if (Status.Get (p) == INSIGNIFICANT)
       {  if (NewStatus == SIGNIFICANT)
          {  if   (Coder.Decode (SIGN_CONTEXT) == 1)
                  Image.Set (p, +1.5 * threshold);
             else Image.Set (p, -1.5 * threshold);
             Status.Set (p, NewStatus);
          }
       }
       else
       {  
          double val = Image.Get (p);
          double delta = threshold * 0.5;
          if (Coder.Decode (REFINEMENT_CONTEXT) == 0) delta = -delta;
          if (val < 0) delta = -delta;
          Image.Set (p, val + delta);
       }
    }
  }
  

  static void main (String[] args)
    throws java.io.IOException
  {
    Signal2D S = new Signal2D (new Position (2, 2));
    S.Set (new Position (0, 0), 15.0);
    S.Set (new Position (1, 0),  8.0);
    S.Set (new Position (0, 1),  7.0);
    S.Set (new Position (1, 1),  1.0);

    Bitstream B = new Bitstream ("hugh");
    HuffmanCoder C = new HuffmanCoder (B);
    QuadTreeCoder Q = new QuadTreeCoder (S);
    Q.Encode (C, 10);
    C.Close ();

    for (int bp = 1; bp <= 10; bp ++)
    {
      B.Reset ();
      S.SetAll (0);

      QuadTreeCoder Q2 = new QuadTreeCoder (S);
      Q2.Decode (C, bp);
      S.Print ();
      System.out.println ();
    }

    if (args.length < 2)
    {  System.out.println ("please enter two PGM-filenames");
       return;
    }

    Signal2D S3 = Signal2D.ReadPgm(args[0]);
    QuadTreeCoder Q3 = new QuadTreeCoder (S3);
    Bitstream B3 = new Bitstream ("hugh");
    HuffmanCoder C3 = new HuffmanCoder (B3);
    Q3.Encode (C3, 3);
    C3.Close ();
    B3.Reset ();
    S3.SetAll (0);
    QuadTreeCoder Q4 = new QuadTreeCoder (S3);
    Q4.Decode (C3, 3);
    S3.WritePgm(args[1]);
  }
}
