class Quantized1D extends Int1DArray
{
  private int Samples[];

  /** create an empty signal with a certain length */
  Quantized1D (int Length)
  {
    Samples = new int[Length];
  }

  Quantized1D (String str)
  {
    Samples = new int[str.length ()];
    for (int x = 0; x < str.length(); x++)
        Samples[x] = (int) str.charAt (x);
  }

  public int Get (int n)
  {
    return Samples[n];
  }

  public void Set (int n, int v)
  {
    Samples[n] = v;
  }

  public int Length ()
  {
    return Samples.length;
  }

  public void Quantize (Double1DArray S, double delta)
  {
    double q = 1.0 / delta;
    for (int i = 0; i < Length (); i ++)
        Samples[i] = (int) Math.round (S.Get (i) * q);
  }

  public void Dequantize (Double1DArray S, double delta)
  {
    for (int i = 0; i < Length (); i ++)
        if   (Samples[i] >= 0)
             S.Set (i, (double) Samples[i] * delta);
        else S.Set (i, (double) Samples[i] * delta);
  }

  public static void main (String[] args)
  {
     Signal1D S = new Signal1D (6);
     Signal1D T = new Signal1D (6);
     Quantized1D Q = new Quantized1D (6);
     S.Set (0, 1.2);
     S.Set (1, 2.5);
     S.Set (2, 3.7);
     S.Set (3, -1.2);
     S.Set (4, -2.5);
     S.Set (5, -3.7);
     System.out.print ("non-quantized: ");
     S.Print ();
     System.out.print ("    quantized: ");
     Q.Quantize (S, 0.5);
     Q.Print ();
     System.out.print (" de-quantized: ");
     Q.Dequantize (T, 0.5);
     T.Print ();
     System.out.print ("MSE = ");
     System.out.print (S.MSE (T));
     System.out.print (", PSNR = ");
     System.out.println (S.PSNR (T, 8.0));
  }

}
