//
//  RLECoder.java
//  jmc
//
//  Created by Dietmar Planitzer on Sat Jun 07 2003.
//  Copyright (c) 2003 __MyCompanyName__. All rights reserved.
//

class RLECoder extends Object
{
   public static final int	MODE = 0;
   public static final int		COMPRESSED = 0;
   public static final int		PLAIN = 1;
   public static final int	LENGTH0 = 1;
   public static final int	SYMBOL0 = 2;
   public static final int	LENGTH1 = 3;
   public static final int	SYMBOL1 = 4;

   
   /**
    * Create a new RLE coder instance.
    */
   RLECoder() {
      super();
   }
   
   /**
    * Encode the given integer array.
    *
    * @param coder coder to write to
    * @param data input data
    */
   void Encode(EntropyCoder coder, Int1DArray data) {   
      // Don't bother with encoding, if nothing to do
      if(data.Length() > 0) {
         int	i = 0, si, imax = data.Length(), c;
         
         while(i < imax) {
            si = i;
            c = data.Get(i++);
            
            // Determine appropriate coding mode
            if(i < imax && data.Get(i) == c) {
               // Compression mode: Scan until we find a symbol which differs from 'c'
               coder.Encode(COMPRESSED, MODE);
               
               // Preflight 'data' in order to find the length of the current run
               while(i < imax) {
                  if(data.Get(i) != c)
                     break;                   
                  i++;
               }
               
               // Write stuff
               coder.Encode(i - si, LENGTH0);
               coder.Encode(c, SYMBOL0);
            } else {
               // Pass-through mode: Scan until we find at least two equal symbols in a row
               coder.Encode(PLAIN, MODE);
               
               // Preflight 'data' in order to find the length of the current run
               while(i < imax) {
                  if(i >= imax || data.Get(i+1) == data.Get(i))
                     break;                  
                  i++;
               }
               
               // Write stuff
               coder.Encode(i - si, LENGTH1);
               coder.Encode(c, SYMBOL1);
               
               i = si+1;
               while(i < imax) {
                  c = data.Get(i);
                  if(i >= imax || data.Get(i+1) == c)
                     break;
                     
                  coder.Encode(c, SYMBOL1);
                  i++;
               }
            }
         }
      }
   }

   /**
    * Same as above, but takes a String instance.
    *
    * @param coder coder to write to
    * @param data input data
    */
   void EncodeString(EntropyCoder coder, String str) {
      GrowableIntArray	data = new GrowableIntArray();
      
      if(str == null)
         throw new IllegalArgumentException();
      
      for(int i = 0; i < str.length(); i++)
         data.Set(i, (int) str.charAt(i));
         
      Encode(coder, data);
   }

   /**
    * Decode the given RLE compressed data.
    *
    * @param coder coder to read from
    * @return the decompressed data
    */
   Int1DArray Decode(EntropyCoder coder) throws java.io.IOException {
      GrowableIntArray	data = new GrowableIntArray();
      int					mode, counter, i = 0, c;

      try {
         for(;;) {
            mode = coder.Decode(MODE);
            
            switch(mode) {
               case COMPRESSED:
                        // Compression mode: Decode `counter' symbols by replication
                        counter = coder.Decode(LENGTH0);
                        c = coder.Decode(SYMBOL0);
                        
                        for(int j = 0; j < counter; j++)
                           data.Set(i++, c);
                        break;
                        
               case PLAIN:
                        // Pass-through mode: Decode `counter' symbols by copying them
                        counter = coder.Decode(LENGTH1);

                        for(int j = 0; j < counter; j++)
                           data.Set(i++, coder.Decode(SYMBOL1));
                        break;
                        
               default:
                        throw new IllegalArgumentException("Malformed input data.");
            }
         }
      } catch(java.io.EOFException ex) {
         // we're done...
      }
      
      return data;
   }
   
   /**
    * Same as above but returns a String instance.
    *
    * @param coder coder to read from
    * @return the decompressed data
    */
   String DecodeString(EntropyCoder coder) throws java.io.IOException {
      Int1DArray		data = Decode(coder);
      StringBuffer	str = new StringBuffer(data.Length());
      
      for(int i = 0; i < data.Length(); i++)
         str.append((char) data.Get(i));
         
      return str.toString();
   }

   
   
   public static void main (String[] args) {
      try {
         Bitstream		stream = new Bitstream("test.bs");
         HuffmanCoder	coder = new HuffmanCoder(stream);
         RLECoder			codec = new RLECoder();
         Int1DArray		data;
         String			str = "abccccddbacbbb";
// HuffmanTable throws an java.lang.ArrayIndexOutOfBoundsException with the following:
//         String			str = "abcd";
      
         System.out.println(str);
         
         // encode sample data
         codec.EncodeString(coder, str);
         coder.Close();
         
         // decode sample data
         stream.Reset();
         str = codec.DecodeString(coder);
         System.out.println(str);
      } catch(Exception ex) {
         ex.printStackTrace();
      }
   }
}
