import java.io.*;

class Signal2D extends Double2DArray
{
  private double Samples[][];

  Signal2D (Position Dim)
  {
    Samples = new double[Dim.Y ()][Dim.X ()];
  }

  double Get (Position p)
  {
    return Samples[p.Y ()][p.X ()];
  }

  void Set (Position p, double v)
  {
    Samples[p.Y ()][p.X ()] = v;
  }

  Position Dimension ()
  {
    return new Position (Samples[0].length, Samples.length);
  }

  void WaveletTransform (Filter Low, Filter High, int Depth)
  {
    int x, y;
    Signal2D signal_low, signal_high;
    DoubleRow dest_row_low, dest_row_high;
    DoubleColumn dest_col_low, dest_col_high;
    int width = Samples[0].length;
    int height = Samples.length;

    while ( Depth-- > 0 ) {
      // horizontale Filterung
      width /= 2;
      signal_low = new Signal2D( new Position(width, 1) );
      signal_high = new Signal2D( new Position(width, 1) );
      dest_row_low = new DoubleRow(signal_low, 0);
      dest_row_high = new DoubleRow(signal_high, 0);
      for (y = 0; y < height; y++) {
        DoubleRow src = new DoubleRow(this, y);
        Low.Apply(src, dest_row_low);
        High.Apply(src, dest_row_high);
        for (x = 0; x < width; x++)
          src.Set( x, dest_row_low.Get(x) );
        for (x = 0; x < width; x++)
          src.Set( x+width, dest_row_high.Get(x) );
      }
      // vertikale Filterung
      height /= 2;
      signal_low = new Signal2D( new Position(1, height) );
      signal_high = new Signal2D( new Position(1, height) );
      dest_col_low = new DoubleColumn(signal_low, 0);
      dest_col_high = new DoubleColumn(signal_high, 0);
      for (x = 0; x < width*2; x++) {
        DoubleColumn src = new DoubleColumn(this, x);
        Low.Apply(src, dest_col_low);
        High.Apply(src, dest_col_high);
        for (y = 0; y < height; y++)
          src.Set( y, dest_col_low.Get(y) );
        for (y = 0; y < height; y++)
          src.Set( y+height, dest_col_high.Get(y) );
      }
    }
  }

  void WaveletReconstruction (Filter Low, Filter High, int Depth)
  {
    int x, y;
    Signal2D signal_low, signal_high;
    DoubleRow src_row_low, src_row_high;
    DoubleColumn src_col_low, src_col_high;
    int width = Samples[0].length >> Depth;
    int height = Samples.length >> Depth;

    while ( Depth-- > 0 ) {
      // vertikale inverse Filterung
      signal_low = new Signal2D( new Position(1, height) );
      signal_high = new Signal2D( new Position(1, height) );
      src_col_low = new DoubleColumn(signal_low, 0);
      src_col_high = new DoubleColumn(signal_high, 0);
      for (x = 0; x < width*2; x++) {
        DoubleColumn dest = new DoubleColumn(this, x);
        for (y = 0; y < height; y++) {
          src_col_low.Set(y, dest.Get(y));
          dest.Set(y, 0.0);
        }
        for (y = 0; y < height; y++) {
          src_col_high.Set( y, dest.Get(y+height) );
          dest.Set(y+height, 0.0);
        }
        Low.InverseApply(src_col_low, dest);
        High.InverseApply(src_col_high, dest);
      }
      height *= 2;
      // horizontale inverse Filterung
      signal_low = new Signal2D( new Position(width, 1) );
      signal_high = new Signal2D( new Position(width, 1) );
      src_row_low = new DoubleRow(signal_low, 0);
      src_row_high = new DoubleRow(signal_high, 0);
      for (y = 0; y < height; y++) {
        DoubleRow dest = new DoubleRow(this, y);
        for (x = 0; x < width; x++) {
          src_row_low.Set(x, dest.Get(x));
          dest.Set(x, 0.0);
        }
        for (x = 0; x < width; x++) {
          src_row_high.Set(x, dest.Get(x+width));
          dest.Set(x+width, 0.0);
        }
        Low.InverseApply(src_row_low, dest);
        High.InverseApply(src_row_high, dest);
      }
      width *= 2;
    }
  }

  //read skipping comments
  private static int Pgm_Read (InputStream s) throws IOException
  {
    int tmp = s.read();
    if(tmp == '#')
        while((tmp != '\n') && (tmp != -1))
            tmp = s.read();

    return tmp;
  }

  //little function to read an unsigned decimal ASCII integer
  private static int Pgm_ReadInt(InputStream s) throws IOException 
  {
      String st = new String("");
      int tmp = ' ';
      while((tmp < '0') || (tmp > '9'))
        tmp = Pgm_Read(s);
    
    while((tmp >= '0') && (tmp <= '9'))
    {    
        st = st.concat(String.valueOf((char)tmp));
        tmp = Pgm_Read(s);
    }
    
    return Integer.parseInt(st);
  }

  /**
   * Opens and reads a PGM file (binary/ASCII)
   * @param FileName the name of the file to be read
   * @return A Signal2D object containing the image data (0.0=black - 1.0=white)
   */
  static Signal2D ReadPgm (String FileName) 
  {
      try
      {
        Signal2D mySignal;
        int width, height, maxval;
        FileInputStream file = new FileInputStream(FileName);
         
        //check magic number
        byte[] mn = new byte[2];
        file.read(mn, 0, 2);
        if((mn[0] != 'P') || ((mn[1] != '2') && (mn[1] != '5')))
            return null;
             
        //parse the header
        width = Pgm_ReadInt(file);
        height = Pgm_ReadInt(file);
        maxval = Pgm_ReadInt(file);
         
        mySignal = new Signal2D(new Position(width, height));
         
        //read the pixels
        for(int y=0; y < height; y++)
            for(int x=0; x < width; x++)
                if(mn[1] == '2') //P2 => ASCII
                    mySignal.Set(new Position(x, y), ((double)Pgm_ReadInt(file)) / (double)maxval);
                else //P5 => binary
                    if(maxval < 256) //1 byte
                        mySignal.Set(new Position(x, y), ((double)file.read()) / (double)maxval);
                      else //2 bytes
                      {
                          int tmp = Pgm_Read(file);
                          tmp = tmp << 8;
                          tmp += Pgm_Read(file);
                        mySignal.Set(new Position(x, y), ((double)tmp) / (double)maxval); 
                      }
               
        file.close();
     
        return mySignal;
      }
      catch(IOException e)
      {
        System.err.print("Signal2D::ReadPgm => ");
          System.err.println(e.getMessage());
          return null;
      }
  }

  /**
   * Exports the current signal into a PGM file
   * @param FileName the name of the file to be written
   */
  void WritePgm (String FileName)
  {
      WritePgm(FileName, false);
  }
  
  /**
   * Exports the current signal into a PGM file (binary/ASCII)
   * @param FileName the name of the file to be written
   * @param ASCII whether to use ASCII (true) or binary (false) format 
   */
  void WritePgm (String FileName, boolean ASCII)
  {
     try
     {
        FileOutputStream file = new FileOutputStream(FileName);
        PrintWriter writer = new PrintWriter(file);
    
        //header
        if(ASCII)
            writer.write("P2\n");
        else
            writer.write("P5\n");
        writer.write("#WritePgm by Martin Heimlich\n");
        writer.print(Dimension().X());
        writer.write(" ");
        writer.print(Dimension().Y());
        writer.write("\n255\n");
        writer.flush();
    
        //data
        for(int y=0; y < Dimension().y; y++)
            for(int x=0; x < Dimension().x; x++)
            {
                if(ASCII)
                {
                    writer.print((int)(255 * Get(new Position(x, y))));
                    writer.write("\n");
                }
                else
                    file.write((int)(255 * Get(new Position(x, y))));
            }
        
        writer.close();
        file.close();
      }
    catch(IOException e)
    {
        System.err.print("Signal2D::WritePgm => ");
        System.err.println(e.getMessage());
    }
  }

  public static void main (String[] args)
  {
     Signal2D S = new Signal2D (new Position (4, 4));
     for (PositionIterator i = new PositionIterator (S.Dimension ()); ! i.AtEnd (); i.Iterate ())
         S.Set (i.Current (), (double) i.Current ().X () * 10.0 + (double) i.Current ().Y () + 11.0);
     S.Print ();

     System.out.println ();
     S.WaveletTransform (Filter.NewLowHaar (), Filter.NewHighHaar (), 2);
     S.Print ();

     System.out.println ();
     S.WaveletReconstruction (Filter.NewLowHaar (), Filter.NewHighHaar (), 2);
     S.Print ();

     if (args.length < 2)
     {  System.out.println ("please specify two PGM-filenames");
        return;
     }

     Signal2D S2 = Signal2D.ReadPgm(args[0]);
     if (S != null)
     {
        for (PositionIterator i = new PositionIterator (S2.Dimension ()); !i.AtEnd (); i.Iterate ())
            S2.Set (i.Current (), S2.Get (i.Current ()) * (double) i.Current().X() / (double) S2.Dimension ().X());
        S2.WritePgm(args[1]);
     }
     else
     {  System.out.print("could not read PGM ");
        System.out.println (args[0]);
     }
  }
}
